

%name logfilterParser
%token_prefix TK_
%token_type { char * }

%token_destructor {
	if($$) free($$);
	if(pargs) {} /* compiler warns for unused variable otherwise */
}

%include {
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <sqlextract/filter/arbval.h>
#include <sqlextract/filter/filter.h>
#include <sqlextract/tables/table.h>
#include <sqlextract/tables/tablespace.h>
#include <time.h> /* for KW_CTIME */
#include <sqlextract/strtools.h>

#include <sqlextract/command.h>
#include <sqlextract/commands/command_select.h>

#if INTERFACE
#define TK_EOF 0
#include <sqlextract/filter/filter.h>
#include <sqlextract/strid.h>
#include <sqlextract/error.h>
#include <sqlextract/util.h>

struct logfilterParserParams {
	struct strid_t *si;
	struct command_t *command;
	struct tablespace_t *tablespace;
	struct table_t *table_singlerow;
	logfilter_error error;
};
#endif
}

%extra_argument { struct logfilterParserParams *pargs }

%parse_accept {
}
%parse_failure {
	pargs->error = ERR_PARSE_FAILURE;
}
%syntax_error {
	pargs->error = ERR_SYNTAX_ERROR;
}

%fallback ID
  ABORT ACTION AFTER ANALYZE ASC ATTACH BEFORE BEGIN BY CASCADE CAST COLUMNKW
  CONFLICT DATABASE DEFERRED DESC DETACH EACH END EXCLUSIVE EXPLAIN FAIL FOR
  IGNORE IMMEDIATE INITIALLY INSTEAD LIKE_KW MATCH NO PLAN
  QUERY KEY OF OFFSET PRAGMA RAISE RECURSIVE REGEXP RELEASE REPLACE RESTRICT ROW
  ROLLBACK SAVEPOINT TEMP TRIGGER VACUUM VIEW VIRTUAL WITH WITHOUT
  REINDEX RENAME CTIME_KW IF
  .
%wildcard ANY.


%left OR.
%left AND.
%right NOT.
%left IS MATCH LIKE_KW REGEXP BETWEEN IN ISNULL NOTNULL NE EQ.
%left GT LE LT GE.
%right ESCAPE.
%left BITAND BITOR LSHIFT RSHIFT.
%left PLUS MINUS.
%left STAR SLASH REM.
%left CONCAT.
%left COLLATE.
%right BITNOT.

// Main command handling
%start_symbol cmd

%token_class ids  ID|STRING.

//////////////////////// Instead of some token classes ////////////////////////

// %token_class id  ID|INDEXED.
%type id {char *}
%destructor id {free($$);}

id(Q) ::= ID(I). {
	if(*I == '\"') {
		Q = strdup_unescape(I);
		free(I);
	} else {
		Q = I;
	}
}
id(Q) ::= INDEXED(I). {
	Q = lf_strdup(I);
}


//////////////////////// The SELECT statement /////////////////////////////////
//

// FIXME: This should be something containing a filter, not a filter itself...
%type cmd {struct command_t *}
%destructor cmd {command_destroy($$);}

cmd ::= select(S).  {
	pargs->command = S;
}


%type select {struct command_t *}
%destructor select {command_destroy($$);}

select(Q) ::= with selectnowith(A). {
	Q = A;
}

%type selectnowith {struct command_t *}
%destructor selectnowith {command_destroy($$);}

selectnowith(Q) ::= oneselect(A). {
	Q = A;
}

%type oneselect {struct command_t *}
%destructor oneselect {command_destroy($$);}

oneselect(Q) ::= SELECT distinct selcollist(COLS) from(T) where_opt(W)
                 groupby_opt having_opt orderby_opt limit_opt(LIMIT). {
	Q = command_select_create(COLS, W, T, LIMIT);
}
// oneselect ::= values. {}

// values ::= VALUES LP nexprlist RP. {}
// values ::= values COMMA LP exprlist RP. {}

// distinct ::= DISTINCT. {}
// distinct ::= ALL. {}

distinct ::= . {
}

%type sclp {struct selcollist_t *}
%destructor sclp {
	selcollist_destroy($$);
}

sclp(Q) ::= selcollist(A) COMMA. {
	Q = A;
}
sclp(Q) ::= . {
	Q = NULL;
}

%type selcollist {struct selcollist_t *}
%destructor selcollist {
	selcollist_destroy($$);
}

selcollist(Q) ::= sclp(L) expr(F) as. {
	Q = selcollist_create(L, F);
}
// selcollist ::= sclp STAR. {}
// selcollist ::= sclp nm DOT STAR. {}

// as ::= AS nm. {}
// as ::= ids. {}
as ::= . {
}

%type from {struct table_t *}
%destructor from {
}

from(Q) ::= . {
	Q = pargs->table_singlerow;
}
from(Q) ::= FROM seltablist(T). {
	Q = T;
}

//stl_prefix ::= seltablist joinop.    {}
stl_prefix ::= . {
}


%type seltablist {struct table_t *}
%destructor seltablist {
	table_destroy($$);
}

seltablist(Q) ::= stl_prefix nm(N) dbnm as indexed_opt
                  on_opt using_opt. {
	Q = tablespace_get(pargs->tablespace, strid_get(pargs->si, N));
	free(N);
}

dbnm ::= . {
}
// dbnm ::= DOT nm. {}

// joinop ::= COMMA|JOIN. {}
// joinop ::= JOIN_KW JOIN. {}
// joinop ::= JOIN_KW nm JOIN. {}
// joinop ::= JOIN_KW nm nm JOIN. {}

// on_opt ::= ON expr. {}
on_opt ::= . {
}

indexed_opt ::= . {}
// indexed_opt ::= INDEXED BY nm. {}
// indexed_opt ::= NOT INDEXED. {}

// using_opt ::= USING LP idlist RP. {}
using_opt ::= . {
}

orderby_opt ::= . {
}
// orderby_opt ::= ORDER BY sortlist. {}

// sortlist ::= sortlist COMMA expr sortorder. {}
// sortlist ::= expr sortorder. {}

// sortorder ::= ASC. {}
// sortorder ::= DESC. {}
// sortorder ::= . {}

groupby_opt ::= . {
}
// groupby_opt ::= GROUP BY nexprlist. {}

having_opt ::= . {
}
// having_opt ::= HAVING expr. {}

%type limit_opt {struct select_limit_t *}
%destructor limit_opt {
	select_limit_destroy($$);
}

limit_opt(LIMIT) ::= . {
	LIMIT = NULL;
}
limit_opt(LIMIT) ::= LIMIT expr(VAL_LIMIT). {
	struct arbval_t *zero = arbval_create_integer(0);
	LIMIT = select_limit_create(VAL_LIMIT, filter_create_constant(zero));
	arbval_destroy(zero);
}
limit_opt(LIMIT) ::= LIMIT expr(VAL_LIMIT) OFFSET expr(VAL_OFFSET). {
	LIMIT = select_limit_create(VAL_LIMIT, VAL_OFFSET);
}
limit_opt(LIMIT) ::= LIMIT expr(VAL_OFFSET) COMMA expr(VAL_LIMIT). {
	LIMIT = select_limit_create(VAL_LIMIT, VAL_OFFSET);
}

//////////////////////// COMMON TABLE EXPRESSIONS ////////////////////////////
// This is just a placeholder
with ::= . {
}

%type nm {char *}
%destructor nm {free($$);}

nm(Q) ::= id(S). {
	Q = S;
}
nm(Q) ::= STRING(S). {
	Q = strdup_unescape(S);
	free(S);
}
// nm ::= JOIN_KW. {}

// idlist ::= idlist COMMA nm. {}
// idlist ::= nm. {}


%type where_opt {struct filter_t *}
%destructor where_opt {filter_destroy($$);}

where_opt(Q) ::= . {
	struct arbval_t *av = arbval_create_integer(1);
	Q = filter_create_constant(av);
	arbval_destroy(av);
}
where_opt(Q) ::= WHERE expr(A). {
	Q = A;
}

/////////////////////////// Expression Processing /////////////////////////////

%type expr {struct filter_t *}
%destructor expr {filter_destroy($$);}

%type term {struct arbval_t *}
%destructor term {arbval_destroy($$);}

expr(Q) ::= term(A). {
	Q = filter_create_constant(A);
	arbval_destroy(A);
}
expr(Q) ::= LP expr(A) RP. {
	Q = A;
}
term(Q) ::= NULL. {
	Q = arbval_create_null();
}
expr(Q) ::= id(I). {
	Q = filter_create_table_column(strid_get(pargs->si, I));
	free(I);
}
// expr ::= JOIN_KW. {}
// expr ::= nm DOT nm. {}
// expr ::= nm DOT nm DOT nm. {}
term(Q) ::= INTEGER|FLOAT|BLOB(A). {
	switch(@A) {
		case TK_INTEGER:
			Q = arbval_create_integer(atol(A));
			break;
		case TK_FLOAT:
			Q = arbval_create_real(atof(A));
			break;
		case TK_BLOB:
			Q = arbval_create_null();
			break;
	}
	free(A);
}
term(Q) ::= STRING(A). {
	char *str = strdup_unescape(A);
	Q = arbval_create_string(str);
	free(str);
	free(A);
}
expr ::= VARIABLE. {
}
// expr ::= expr COLLATE ids. {}
// expr ::= id LP distinct exprlist RP. {}
// expr ::= id LP STAR RP. {}

term(Q) ::= CTIME_KW. {
	Q = arbval_create_integer((long)time(NULL));
}

expr(Q) ::= expr(L) AND expr(R). {
	Q = filter_create_and(2, L, R);
}
expr(Q) ::= expr(L) OR expr(R). {
	Q = filter_create_or(2, L, R);
}
expr(Q) ::= expr(L) LT|GT|GE|LE(OP) expr(R). {
	switch(@OP) {
		case TK_LT: Q = filter_create_comp_lt(L,R); break;
		case TK_GT: Q = filter_create_comp_gt(L,R); break;
		case TK_GE: Q = filter_create_comp_ge(L,R); break;
		case TK_LE: Q = filter_create_comp_le(L,R); break;
		default:
			/* Some really strange error */
			Q = filter_create_and(0);
			filter_destroy(L);
			filter_destroy(R);
	}
	free(OP);
}
expr(Q) ::= expr(L) EQ|NE(OP) expr(R). {
	Q = filter_create_comp_eq(L,R);
	if(@OP == TK_NE) {
		Q = filter_create_not(Q);
	}
	free(OP);
}
expr ::= expr BITAND|BITOR|LSHIFT|RSHIFT expr. {
}
expr(Q) ::= expr(L) PLUS expr(R). {
	Q = filter_create_add(L,R);
}
expr(Q) ::= expr(L) MINUS expr(R). {
	Q = filter_create_sub(L,R);
}

expr(Q) ::= expr(L) STAR|SLASH|REM(OP) expr(R). {
	Q = NULL;
	switch(@OP) {
		case TK_STAR:  Q = filter_create_mul(L, R); break;
		case TK_SLASH: Q = filter_create_div(L, R); break;
		case TK_REM:   Q = filter_create_mod(L, R); break;
	}
	free(OP);
}
expr(Q) ::= expr(L) CONCAT expr(R). {
	Q = filter_create_concat(L, R);
}

%type regexpop {int}
%destructor regexpop {}
regexpop(Q) ::= REGEXP(OP). {
	Q = FILTER_STRINGS_CASE_INSENSITIVE;
	free(OP);
}
regexpop(Q) ::= NOT REGEXP(OP). {
	Q = FILTER_STRINGS_CASE_INSENSITIVE | FILTER_STRINGS_INVERSE;
	free(OP);
}

/* The following rule is not part of sqlite, but extension for logfilter */ 
regexpop(Q) ::= REGEXP(OP) BINARY. {
	Q = FILTER_STRINGS_NONE;
	free(OP);
}

/* The following rule is not part of sqlite, but extension for logfilter */
regexpop(Q) ::= NOT REGEXP(OP) BINARY. {
	Q = FILTER_STRINGS_INVERSE;
	free(OP);
}

expr(Q) ::= expr(L) regexpop(FLAGS) expr(R). [REGEXP] {
	Q = filter_create_regexp(&(pargs->error), L, R, FLAGS);
}

%type likeop {int}
%destructor likeop {}

likeop(Q) ::= LIKE_KW|MATCH(OP). {
	Q = 0;
	free(OP);
}
likeop(Q) ::= NOT LIKE_KW|MATCH(OP). {
	Q = 1;
	free(OP);
}

expr(Q) ::= expr(L) likeop(NEG) expr(R).  [LIKE_KW] {
	Q = filter_create_like(&(pargs->error), L, R, NULL);
	if(NEG) {
		Q = filter_create_not(Q);
	}
}
expr(Q) ::= expr(L) likeop(NEG) expr(R) ESCAPE expr(E).  [LIKE_KW] {
	Q = filter_create_like(&(pargs->error), L, R, E);
	if(NEG) {
		Q = filter_create_not(Q);
	}
}

expr(Q) ::= expr(L) ISNULL|NOTNULL(OP). {
	struct arbval_t *null = arbval_create_null();
	
	Q = filter_create_comp_is(L, filter_create_constant(null));
	
	if(@OP == TK_NOTNULL) {
		Q = filter_create_not(Q);
	}
	
	arbval_destroy(null);
	free(OP);
}
expr(Q) ::= expr(L) NOT NULL. {
	struct arbval_t *null = arbval_create_null();
	Q = filter_create_not(filter_create_comp_is(L, filter_create_constant(null)));
	arbval_destroy(null);
}

expr(Q) ::= expr(L) IS expr(R). {
	Q = filter_create_comp_is(L, R);
}

expr(Q) ::= expr(L) IS NOT expr(R). {
	Q = filter_create_not(filter_create_comp_is(L, R));
}



expr(Q) ::= NOT expr(A). {
	Q = filter_create_not(A);
}
expr(Q) ::= BITNOT expr(A). {
	/*
	 * Should be the bit inverse value, which is actually -1-x, since integers are
	 * 2's complement values for negative values.
	 */
	struct arbval_t *negone = arbval_create_integer(-1);
	Q = filter_create_sub( filter_create_constant(negone), A );
	arbval_destroy(negone);
}
expr(Q) ::= MINUS expr(A). [BITNOT] {
	struct arbval_t *zero = arbval_create_integer(0);
	Q = filter_create_sub( filter_create_constant(zero), A );
	arbval_destroy(zero);
}
expr(Q) ::= PLUS expr(A). [BITNOT] {
	Q = A;
}

%type between_op {int}
%destructor between_op {}

between_op(Q) ::= BETWEEN. {
	Q = 0;
}
between_op(Q) ::= NOT BETWEEN. {
	Q = 1;
}
expr(Q) ::= expr(VAL) between_op(NEG) expr(LOW) AND expr(HIGH). [BETWEEN] {
	Q = filter_create_between(VAL, LOW, HIGH);
	if(NEG) {
		Q = filter_create_not(Q);
	}
}

/* CASE expressions */

// expr ::= CASE case_operand case_exprlist case_else END. {}
// case_exprlist ::= case_exprlist WHEN expr THEN expr. {}
// case_exprlist ::= WHEN expr THEN expr. {}
// case_else ::=  ELSE expr. {}
// case_else ::=  . {}
// case_operand ::= expr. {}
// case_operand ::= . {}

// exprlist ::= nexprlist. {}
// exprlist ::= . {}
// nexprlist ::= nexprlist COMMA expr. {}
// nexprlist ::= expr. {}
