#include <sqlextract/tables/table.h>
#include <sqlextract/filter/filter.h>
#include <sqlextract/util.h>
#include <stdlib.h>

struct table_t {
	const struct table_desc_t *tdesc;
	void *table_storage;
};

struct table_row_t {
	struct table_t *tbl;
	void *iterator_storage;
};

struct filter_table_column_storage_t {
	int fieldname;
};

struct table_t *table_create(const struct table_desc_t *tdesc,
		struct strid_t *si, void *arg) {
	struct table_t *tbl = lf_calloc(1, sizeof(struct table_t));
	tbl->tdesc = tdesc;
	tbl->table_storage = (*tbl->tdesc->create_table)(si, arg);
	return tbl;
}

void table_destroy(struct table_t *tbl) {
	if (tbl == NULL)
		return;
	if (tbl->tdesc->destroy_table)
		(*tbl->tdesc->destroy_table)(tbl->table_storage);
	free(tbl);
}

void table_filter(struct table_t *tbl, int (*cb)(struct table_row_t *, void *),
		void *arg, struct filter_t *flt) {
	struct table_row_t *row = lf_calloc(1, sizeof(struct table_row_t));
	int running;

	row->tbl = tbl;
	row->iterator_storage = (*tbl->tdesc->create_iterator)(tbl->table_storage);

	while ((*tbl->tdesc->fetch_row)(row->iterator_storage)) {
		if (arbval_is_true(filter_process(flt, row))) {
			running = (*cb)(row, arg);
			if (!running)
				break;
		}
	}

	(*tbl->tdesc->destroy_iterator)(row->iterator_storage);
	free(row);
}

struct arbval_t *table_row_fetch_column(struct table_row_t *row, int fieldname) {
	struct table_t *tbl = row->tbl;
	return (*tbl->tdesc->fetch_field)(row->iterator_storage, fieldname);
}

static void filter_destroy_table_column(void *storage) {
	struct filter_table_column_storage_t *stg =
			(struct filter_table_column_storage_t *) storage;
	free(stg);
}

static void filter_process_table_column(struct arbval_t *result, void *object,
		void *storage) {
	struct filter_table_column_storage_t *stg =
			(struct filter_table_column_storage_t *) storage;
	struct table_row_t *row = (struct table_row_t *) object;
	if(row == NULL) {
		arbval_typecast_null(result);
	} else {
		arbval_clone_into(result, table_row_fetch_column(row, stg->fieldname));
	}
}

struct filter_t *filter_create_table_column(int fieldname) {
	struct filter_table_column_storage_t *stg = lf_calloc(1,
			sizeof(struct filter_table_column_storage_t));

	stg->fieldname = fieldname;

	return filter_create(filter_process_table_column,
			filter_destroy_table_column, (void*) stg);
}
